#!/usr/bin/perl

#
# dr_align.pm: some underlying functions that should be useful to all
#              implementations of dr_align
# 
#
# $Header: /u/drspeech/src/dr_scripts/RCS/dr_align.pm,v 1.5 2000/05/09 22:43:31 fosler Exp $
#

package dr_align;
require Exporter;
@ISA = qw(Exporter);
 
@EXPORT = qw(get_next_utt phi2phndict ctm_fill_empty_utts);  # symbols to export by default
@EXPORT_OK = qw();    # symbols to export on request
%EXPORT_TAGS = ();  # define names for sets of symbols

sub get_next_utt {
    my $fh=shift;
    my $filetype=shift;
    my ($tag,$chan,$spkr,$start,$stop,$cond,$ref);
    my $id;

    if (!defined($fh)) {
	$fh=\*STDIN;
    }
    if (!defined($filetype)) {
	$filetype="ref";
    }

    $ref=undef;
    $id=undef;

    while($line=<$fh>) {
	if (($filetype eq "stm" && $line=~/^;;/) ||
	    (($filetype eq "ref" || $filetype eq "tagref") && $line=~/^\#/)) {
	    next;
	}
	if ($filetype eq "stm") {
	    ($tag,$chan,$spkr,$start,$stop,$cond,$ref)=split(/[ 	]+/,$line,7);
	    $id=$tag."_".int($start*100)."_".int($stop*100);
	    $ref=~s/[(][%]HESITATION[)]//g;
	    $ref=~s/  */ /g;
	} elsif ($filetype eq "ref") {
	    $line=~s/[ 	]+$//;
	    # check for ids
	    if ($line =~ /^[0-9\(]/) {
		($id,$ref)=split(/[ 	]+/,$line,2);
		$id=~s/[()]//g;
	    } elsif ($line =~ /[\)]$/) {
		($ref,$id)=($line=~/^(.*)[ 	]+([^ 	]+)/);
		$id=~s/[()]//g;
	    } else {
		$ref=$line;
		$id=undef;
	    }
	} elsif ($filetype eq "tagref") { # we know that first word is uttid
	    $line=~s/[ 	]+$//;
	    ($id,$ref)=split(/[ 	]+/,$line,2);
	} else {  # who knows?
	    warn("dr_align::get_next_utt: unknown filetype");
	    $ref=$line;
	    $id=undef;
	}
	last;
    }

    if (defined($ref)) {
	return wantarray ? ($ref,$id) : $ref;
    } else {
	return wantarray ? () : undef;
    }
}


# converts a phi file into a phone dictionary, and returns the number of phones

sub phi2phndict {
    my $phi=shift;
    my $phndict=shift;
    my @phones=();
    my @x;
    my $p;

    open(PHI,"$phi") || die("dr_align_efsg::phi2phndict: Can't open $phi for reading");

    $_=<PHI>;
    chop;
    if ($_ ne "PHI") {
	die("dr_align_efsg::phi2phndict: bad magic number (phi file $phi doesn't start with PHI)");
    }
    while(<PHI>) {
	next if /^\#/;
	last;
    }
    chop;
    $nphones=$_;
    while(<PHI>) {
	next if (/^\#/);
	@x=split;
	push(@phones,$x[0]);
    }
    if ($#phones+1 != $nphones) {
	$nxphones=$#phones+1;
	die("dr_align_efsg::phi2phndict: number of phones in file ($nxphones) != listed number of phones ($nphones)");
    }
    close(PHI);
    open(PHNDICT,">$phndict") || die("dr_align_efsg::phi2phndict: Can't open $phndict for writing");
    push(@phones,"NULL");
    foreach $p (sort @phones) {
	if ($p eq "NULL") {
	    print PHNDICT ("NULL(1.0)\n");
	} else {
	    printf PHNDICT ("%s(1.0) %s\n",$p,$p);
	}
    }
    close(PHNDICT);
    return $nphones;
}

# checks for bad (unaligned) utterances in a CTM file, and fills 
# with the EMPTY phone.  Checks against pfile.

# usage: &ctm_fill_empty_utts(inctm,outctm,pfile,framestepms,zeropadms,[baduttslistref],[overwriteok]);
sub ctm_fill_empty_utts {
    my $inctm=shift;
    my $outctm=shift;
    my $pfile=shift;
    my $framestep=shift;   # in milliseconds
    my $zeropad=shift;     # in milliseconds
    my $baduttslist=shift;
    my $overwriteok=shift;
    my $uttskip=shift;

    my ($b,$num,$numframes,$ctmline,$uttnum,$lastuttnum,$rest,$starttime,$endtime,$length,$channel);
    my @badutts;
    my @framecount;
    my @ctmbuf;
    my $line;
    
    
    die("dr_align::ctm_fill_empty_utts: inctm not defined") if !defined($inctm);
    die("dr_align::ctm_fill_empty_utts: outctm not defined") if !defined($outctm);
    die("dr_align::ctm_fill_empty_utts: pfile not defined") if !defined($pfile);
    die("dr_align::ctm_fill_empty_utts: framestep not defined") if !defined($framestep);
    die("dr_align::ctm_fill_empty_utts: zeropad not defined") if !defined($zeropad);

    if (defined($baduttslist)) {
	foreach $b (@{$baduttslist}) {
	    $badutts[$b]=1;
	}
    }

    if (!defined($uttskip)) {
	$uttskip=0;
    }

    if (-f $outctm && !$overwriteok) {
	die("dr_align::ctm_fill_empty_utts: output file $outctm exists and overwrite permission not set");
    }
    
    if ($drspeech::verbose >=2) {
	&drspeech::inform("dr_align::ctm_fill_empty_utts: converting $inctm to $outctm");
	if ($drspeech::verbose >=3) {
	    &drspeech::inform("\tpfile=$pfile");
	    &drspeech::inform("\tframestep=$framestep");
	    &drspeech::inform("\tzeropad=$zeropad");
	    &drspeech::inform("\tbaduttslist=".join(",",@{$baduttslist}));
	    &drspeech::inform("\toverwriteok=$overwriteok");
	}
    }

    # get the list of how many frames in each utt
    open(PFILE,"pfile_info -i $pfile -p -q |") || die("can't open pipe on $pfile");
    while(<PFILE>) {
	chop;
	($num,$nframes)=split;
	if ($num < $uttskip) {
	    next;
	}
	# CTM files are one-indexed, so put the pfile number into $num+1
	$framecount[$num+1-$uttskip]=$nframes-2*(int(0.5+$zeropad/$framestep));
    }
    close(PFILE);

    &drspeech::inform("Done reading pfile $pfile") if ($drspeech::verbose>=3);

    my $framestepsec=$framestep/1000;
    my $utt;

    open(INCTM,"$inctm") || die("dr_align::ctm_fill_empty_utts: can't open $inctm for reading");
    open(OUTCTM,">$outctm") || die("dr_align::ctm_fill_empty_utts: can't open $outctm for writing");
    $lastuttnum=0;
    while(<INCTM>) {
	chop;
	$ctmline=$_;
	($uttnum,$channel,$starttime,$length,$rest)=split;
	if ($uttnum!=$lastuttnum && $lastuttnum>=0) {
	    #check out lastutt
	    if ($badutts[$lastuttnum] && $lastuttnum>0) {
		
		printf OUTCTM ("%d A %.3f %.3f EMPTY+BAD_ALIGN\n",
			       $lastuttnum, 0, 
			       $framecount[$lastuttnum]*$framestepsec);
	    } elsif ($lastuttnum>0) {
		if (int(0.5+(2*$zeropad/1000+$endtime)/$framestepsec) != $framecount[$lastuttnum]) {
		    my $ct=int(0.5+(2*$zeropad/1000+$endtime)/$framestepsec);
		    warn("dr_align::ctm_fill_empty_utts: utterance $lastuttnum has different number of frames in CTM file ($ct) and pfile ($framecount[$lastuttnum])");
		}
		foreach $line (@ctmbuf) {
		    print OUTCTM $line,"\n";
		}
	    }
	    @ctmbuf=();
	    #check for skipped sentences
	    for ($utt=$lastuttnum+1;$utt<$uttnum;$utt++) {
		warn("CTM file $inctm missing utterance $utt, but not in badlist") if (!$badutts[$utt]);
		&drspeech::inform("replacing utt $utt")	if 
		    ($drspeech::verbose>=3);
		printf OUTCTM ("%d A %.3f %.3f EMPTY+BAD_ALIGN\n",
			       $utt, 0, 
			       $framecount[$utt]*$framestepsec);
	    }		
	}
	$lastuttnum=$uttnum;
	$endtime=$starttime+$length;
	push(@ctmbuf,$ctmline);
    }
    
    #process last sentence
    if ($badutts[$lastuttnum]) {
	printf OUTCTM ("%d A %.3f %.3f EMPTY+BAD_ALIGN\n",
		       $lastuttnum, 0, 
		       $framecount[$lastuttnum]*$framestepsec);
    } else {
	if (int(0.5+(2*$zeropad/1000+$endtime)/$framestepsec) != $framecount[$lastuttnum]) {
	    my $ct=int(0.5+(2*$zeropad/1000+$endtime)/$framestepsec);
	    warn("dr_align::ctm_fill_empty_utts: utterance $lastuttnum has different number of frames in CTM file ($ct) and pfile ($framecount[$lastuttnum])");
	}
	foreach $line (@ctmbuf) {
	    print OUTCTM $line,"\n";
	}
    }
    #check for missing final utts
    $lastuttnum++;

    while ($badutts[$lastuttnum]) {
	&drspeech::inform("replacing utt $lastuttnum")	if 
	    ($drspeech::verbose>=3);
	printf OUTCTM ("%d A %.3f %.3f EMPTY+BAD_ALIGN\n",
		       $lastuttnum, 0, 
		       $framecount[$lastuttnum]*$framestepsec);
	$lastuttnum++;
    }
    close(INCTM);
    close(OUTCTM);
}
	
    
1;
