#!/usr/local/bin/perl5
 
#
# efsg.pm: utility routines for efsg building
#
# Written by J. Eric Fosler-Lussier (fosler@icsi.berkeley.edu)
#
# $Header: /u/drspeech/src/lexicon/grapheme2phoneme/gen_efsg/RCS/efsg.pm,v 1.2 1998/06/23 21:39:17 fosler Exp $

package efsg;
use drspeech;

sub find_fragments {
    my $w=shift;
    my $dictwords=shift;
    my $fragment_samplect=shift;
    my @tr=undef;
    my @trialwords=();
    my @trialwordstrings=();
    my $oldw;
    my $i;
    my $x;
    my $ww;
    my $beginfragment;

    $oldw=$w;
    if ($w=~/^[^-]+-$/) {
	$w=~s/-//;
	@tr=grep(/^$w/,@{$dictwords});
	$beginfragment=1;
    } elsif ($w=~/^-[^-]+$/) {
	$w=~s/-//;
	@tr=grep(/$w$/,@{$dictwords});
	$beginfragment=0;
    }
    if (defined(@tr)) {
	for($i=0;$#tr>=0 && $i<$fragment_samplect;$i++) {
	    $x=int(rand($#tr+1));
	    $ww=splice(@tr,$x,1);
	    push(@trialwordstrings,$ww);
	    push(@trialwords,[&word2letters($ww)]);
	}
	if ($drspeech::verbose>=3) {
	    printf STDERR ("Using fragment generator for %s with the folliwing words:\n%s\n",$oldw,join(" ",@trialwordstrings));
	}
	
    } 
    if ($#trialwords>=0) {
	return ($beginfragment,@trialwords);
    } else {
	return ($beginfragment,undef);
    }
}

sub dict2fsg {
    use onward_fsg;
    my $w=shift;
    my $currentstate=0;
    my $dictionary=shift;

    my $total=0;
    my @secondstate=();
    my @dictentries;
    my $fsg=new onward_fsg();
    my $i;

    @dictentries=@{$dictionary->{$w}};

    for ($i=0;$i<=$#dictentries;$i++) {
	$secondstate[$i]=$currentstate+$total+1;
	$total+=($#{$dictentries[$i]}-1);
    }
    $finalstate=$currentstate+$total+1;
    for ($i=0;$i<=$#dictentries;$i++) {
	$fsg->addtrans($currentstate,
		       ($#{$dictentries[$i]}==1)?$finalstate:$secondstate[$i],
		       $dictentries[$i]->[0],
		       $dictentries[$i]->[1]);
    }
    
    $currentstate++;
    for ($i=0;$i<=$#dictentries;$i++) {
	@entry=@{$dictentries[$i]};
	for($j=2;$j<$#entry;$j++) {
	    $fsg->addtrans($currentstate,
			   $currentstate+1,
			   1.0,
			   $entry[$j]);
	    $currentstate++;
	}
	if ($#entry>1) {
	    $fsgs[$wd]->addtrans($currentstate,
				 $finalstate,
				 1.0,
				 $entry[$j]);
	    $currentstate++;
	}
    }
    return $fsg;
}
    

sub letters2fsg {
    use onward_fsg;
    my $w=shift;
    my $curstate=0;
    my $dictwords=shift;
  
    my $trialwords;
    my $beginfragment;
    my @dist;
    my $maxout;
    my $e;
    my @a;
    my @latters;
    my $fsg=new onward_fsg();

    #look for word fragments
    undef $trialwords;
    if (defined($dictwords) && $w=~/-/) {
	($beginfragment,$trialwords)=
	    &find_fragments($w,$dictwords,$fragment_samplect);
    }
    $w=~s/-//;

    @letters=&word2letters($w);

    for($i=3;$i<=$#letters-3;$i++) {
	print "$letters[$i]: " if $drspeech::verbose>=2;
		
	if (!defined($trialwords)) { #just feed the original word in
	    $e=&evaluate($topnode{$letters[$i]},@letters[$i-3..$i+3]);
	    print $e,":" if $drspeech::verbose>=2;
	    @dist=@{$dist{$e}};
	} else {
	    @dist=&evaluate_fragment($topnode{$letters[$i]},$i,$beginfragment,$trialwords,$#letters);
	}
	$maxout=0;
	for($j=0;$j<=$#dist;$j++) {
	    if ($drspeech::verbose>=2) {
		@a=@{$dist[$j]};
		printf("%s/%.6f ",join("_",@a[1..$#a]),$a[0]) if $drspeech::verbose>=2;
	    }
	    $maxout=$#{$dist[$j]} if $#{$dist[$j]}>$maxout;
	}
	print "\n" if $drspeech::verbose>=2;
	$nextstate=$curstate+$maxout;
	for($j=0;$j<=$#dist;$j++) {
	    @out=@{$dist[$j]};
	    $fsg->addtrans($curstate, 
			   ($#out==1)?$nextstate:$curstate+1,
			   $out[0],$out[1]);
	    if ($#out>1) {
		$xstate=$curstate+1;
		for($k=2;$k<=$#out;$k++) {
		    $fsg->addtrans($xstate,$xstate+1,
				   1.0,$out[$k]);
		    $xstate++;
		}
	    }
	}
	$curstate=$nextstate;
    }	
#   $fsgs[$wd]=$fsgs[$wd]->prune_unreachable();
    return $fsg->rmepsilon("NULL");
}   
    
    
# utility routines
sub array2assoc {
    local $c={};
 
    foreach $x (@_) {
        $c->{$x}=0;
    }
    return $c;
}

sub countdepth {
    local $counter=0;
    if ($_[0] =~ /^[0-9\.]+$/) {
        return -1;
    }
 
    while($_[0] eq "|") {
        shift(@_);
        $counter++;
    }
    return $counter;
}

sub renormalize {
    local @a=@_;
    local $total=0;
    local $a;

    foreach $a (@a) {
	$total+=$a;
    }
    foreach $a (@a) {
	$a=$a/$total;
    }
    return @a;
}

sub word2letters {
    my $word=shift;
    my @letters;
    my $l;

    @letters=split("",$word);
    foreach $l (@letters) {
	$l=$efsg_lettermap{$l} if defined $efsg_lettermap{$l};
    }
    unshift(@letters,"HASH","HASH","HASH");
    push(@letters,"HASH","HASH","HASH");
    return @letters;
}	

%efsg_lettermap=(".","DOT",
		 "'","APOST",
		 "#","HASH",
		 "(","LPAREN",
		 ")","RPAREN",
		 "-","DASH",
		 "<","LTHAN",
		 ">","GTHAN",
		 "_","UNDERSCORE",
		 "","a2",
		 "","a3",
		 "","c5",
		 "","e2",
		 "","e1",
		 "","e3",
		 "","e4",
		 "","i3",
		 "","i4",
		 "","i3",
		 "","u2",
		 "","u3",
		 "","u4");
		 
1;
